float EXPFACTOR = 40000.0;

int tilesize = 32;
int fieldMargin = tilesize * 2;

int maxBads = 8;
int minSpawnDis = 7;
float secondsToSpawn = 0.5;
float secondsToMove = 1;

float fadeSeconds = 2;

int score = 0;

int totalTime = 0;

boolean[] dirs = {false, false, false, false}; //up down left right

PFont font;

void newGame()
{
  //badguys
  badguys = new ArrayList();
 
  player = new sprite();
  player.visible = true;
  player.alive = true;
  player.position.x = floor(tilesw/2);
  player.position.y = floor(tilesh/2);
  
  target = new sprite();
  target.visible = target.alive = true;
  target.position = new PVector(player.position.x, player.position.y);
  
  nextMoveThing = 0;
  nextSpawnThing = 0;
  totalTime = 0;
  
  score = 0;
}

void setup()
{
  size(800, 640);
  
  frameRate(30);
  
  //dimensions
  fieldw = width - fieldMargin * 2;
  fieldh = height - fieldMargin * 2;
  tilesw = fieldw / tilesize;
  tilesh = fieldh / tilesize;

  newGame();
  
  smooth();
  
  font = loadFont("Centaur-48.vlw");
  textFont(font);
}

int fieldw;
int fieldh;
int tilesw, tilesh;

class sprite
{
  boolean visible = false;
  boolean alive = true;
  PVector position = new PVector(0,0);
  
  float alph = 255;
  int whendied = 0;
  
  void kill()
  {
    alive = false;
    alph = 255;
    whendied = totalTime;
  }
}


sprite player;
sprite target;
ArrayList badguys;

void draw()
{
  if(focused && player.alive)
  {
    //if not "paused"
    //though my input state machine probably has little hope
    movePlayer();
    moveSprites();
    spawnStuff();
    
    updateTime();
  }
  
  drawEverything();
  
  if(!player.alive)
  {
    drawLose();
  }
  
  if(!focused)
  {
    drawFocus();
  }
 
  resetKeys();
}

void updateTime()
{
  totalTime += (int) (1000/frameRate);
}

int nextMoveThing = 0;

void moveSprites()
{
  if(totalTime >= nextMoveThing)
  {
    //each sprite moves towards the goal!
    for(int i = 0; i < badguys.size(); i++)
    {
      sprite bad = (sprite)badguys.get(i);
      
      if(!bad.alive)
        continue;
      //calculate next space.
      //if another sprite is there, no worries. If the player is there, don't move.
      //determine direction
      PVector direction = new PVector(target.position.x, target.position.y);
      direction.sub(bad.position);
      PVector newPosition = new PVector(bad.position.x, bad.position.y);
      if(abs(direction.x) > abs(direction.y))
      {
        if(direction.x > 0)
          newPosition.x++;
        else
          newPosition.x--;
      }
      else
      {
        if(direction.y > 0)
          newPosition.y++;
        else
          newPosition.y--;
      }
      
      if(newPosition.x == player.position.x && newPosition.y == player.position.y)
      {
        //do nothing -- this would trample the player
      }
      else
      {
        bad.position = new PVector(newPosition.x, newPosition.y);
        
        if(bad.position.x == target.position.x && bad.position.y == target.position.y)
        {
          //GAME OVER!!
          player.kill();
        }
      }
    }
    
    float factor = 1000 * (secondsToMove + 2*exp(- totalTime / EXPFACTOR));
    nextMoveThing = (int)(totalTime + factor);
  }
}

int nextSpawnThing = 0;

sprite getSpriteAt(PVector p)
{
  //returns first badguy at this point
  for(int i = 0; i < badguys.size(); i++)
  {
    sprite s = (sprite) badguys.get(i);
    if(s.position.x == p.x && s.position.y == p.y)
      return s;
  }
  
  return null;
}

void spawnStuff()
{
//  int maxbadsEffective = floor(maxBads * exp(totalTime / (2*EXPFACTOR)) );
  
  if(badguys.size() < maxBads)
  {
    if(totalTime >= nextSpawnThing)
    {
      //enough time has passed
      //decide somehow if this should spawn
      //if(random(0, 1.0) < 0.5)
      {
        //spawn something

        //minimum distance from player, of course
        PVector p = new PVector(target.position.x, target.position.y);
        while(p.dist(target.position) < minSpawnDis
              || getSpriteAt(p) != null)
        {
          p = new PVector(floor(random(0, tilesw)), floor(random(0, tilesh)));
        }
        
        sprite bad = new sprite();
        bad.visible = true;
        bad.alive = true;
        bad.position = p;
        badguys.add(bad);
      }
      
      //in any case ..
      float factor = 1000 * (secondsToSpawn + exp(- totalTime / EXPFACTOR));
      nextSpawnThing = (int) (totalTime +  factor );
    }
  }
}

void drawEverything()
{
  background(0);

  drawScreen();

  //make the field start on the field, graphically too
  pushMatrix();
    translate(fieldMargin, fieldMargin);
    drawSprites();
  popMatrix();
  
  drawScore();
}

void drawScore()
{
  noStroke();
  fill(255);
  textAlign(LEFT);
  text("SCORE: " + score, 5, 48); 
}

void movePlayer()
{
  PVector trymove = new PVector(0,0);
  if(dirs[0])
  {
    trymove.y = -1;
  }
  else if (dirs[1])
  {
    trymove.y = +1;
  }
  else if (dirs[2])
  {
    trymove.x = -1;
  }
  else if (dirs[3])
  {
    trymove.x = 1;
  }

  PVector newPlayerPos = new PVector(trymove.x, trymove.y);  
  newPlayerPos.add(player.position);
  if(newPlayerPos.x < 0 || newPlayerPos.x >= tilesw)
    return;
  if(newPlayerPos.y < 0 || newPlayerPos.y >= tilesh)
    return;
    
  player.position.x = newPlayerPos.x;
  player.position.y = newPlayerPos.y;

  sprite badone;
  while((badone = getSpriteAt(player.position)) != null)
  {
    badone.position.add(trymove);
    //test if enemy fell off an edge!
    if((badone.position.x < 0 || badone.position.x >= tilesw)
      || badone.position.y < 0 || badone.position.y >= tilesh)
      {
        //KILL KILL KILL!!!
        badone.kill();
        float scorefactor = exp( totalTime / EXPFACTOR );
        score += floor(10 * (scorefactor));
      }
  }
}

void resetKeys()
{
  for(int i = 0; i < dirs.length; i++)
  {
    dirs[i] = false;
  }
}

void drawScreen()
{
  //screen size
  color fieldColor = color(0, 255, 0);
  fill(fieldColor);
  noStroke();
  rect(fieldMargin, fieldMargin, fieldw, fieldh);
  //TODO: make nice background stuff!!
}

void keyPressed() {
  if (key == CODED) {
    switch(keyCode)
    {
      case UP:
        dirs[0] = true;
        break;
      case DOWN:
        dirs[1] = true;
        break;
      case LEFT:
        dirs[2] = true;
        break;
      case RIGHT:
        dirs[3] = true;
        break;
    }
  }
  else
 {
   if(key == 'r' || key == 'R')
   {
     newGame();
   }
 } 
}

void drawLose()
{
  stroke(0);
  fill(0);
  textAlign(CENTER);
  text("Game over! Score: " + score, width/2, height/2);
  text("Press R to restart", width/2, height/2 + 50);
}

void drawFocus()
{
  stroke(0);
  fill(0);
  textAlign(CENTER);
  text("[click to focus]", width/2, height/2 - 50);
}

void drawSprites()
{
  ellipseMode(CORNER);
  
  //target
  fill(color(0, 244, 255));
  stroke(color(4, 20, 21));
  rect(target.position.x * tilesize, target.position.y * tilesize, 1, 2*tilesize/3);
  rect(target.position.x * tilesize, target.position.y * tilesize, tilesize/2, tilesize/3);
  noStroke();

  //player  
  fill(color(0, 0, 255));
  ellipse(player.position.x * tilesize, player.position.y * tilesize, tilesize, tilesize);
  
  //badguys
  for(int i = badguys.size()-1; i >= 0; i--)
  {
    sprite badg = (sprite)badguys.get(i);
    fill(color(255, 0, 0, badg.alph));
    rect(badg.position.x * tilesize, badg.position.y * tilesize, tilesize, tilesize);
    if(!badg.alive)
    {
      //reduce alpha to fade out
      badg.alph = floor(map((totalTime - badg.whendied), 0, fadeSeconds * 1000, 255, 0));
      if(badg.alph <= 0)
      {
        badguys.remove(i);
      }
    }
  }
}
